# Beyonic API Python bindings

# Default values if any
DEFAULT_ENDPOINT_BASE = 'https://app.beyonic.com/api/'

# Config
api_key = None
api_endpoint_base = None
api_version = None
verify_ssl_certs = True  # Set to False if you want to bypass SSL checks (mostly useful while testing it on local env).


from beyonic.apis.payment import Payment
from beyonic.apis.webhook import Webhook
from beyonic.apis.collection import Collection
from beyonic.apis.collectionrequest import CollectionRequest
from beyonic.apis.account import Account
from beyonic.apis.contact import Contact
from beyonic.apis.transaction import Transaction
from beyonic.apis.network import Network
from beyonic.apis.currency import Currency

__version__ = "0.1.14"



"""
# This return the api client. It should first check if requests is installed, if its installed then returns RequestsClient,
# if requests is not installed then check for urlfetch lib, if its installed then return its client UrlfetchClient
# if both of them not installed then throws the exception
"""

import textwrap
from beyonic.errors import BeyonicError, ResponseError
from beyonic.resources import GenericObject
import os
import beyonic

# Let's try to import requests if its available
try:
    import requests
except ImportError:
    requests = None

# Let's try to import urlfetch
try:
    from google.appengine.api import urlfetch
except ImportError:
    urlfetch = None


class BaseClient(object):
    def __init__(self, verify_ssl_certs=True):
        self._verify_ssl_certs = verify_ssl_certs

    def transform_params_metadata(self, params):
        # transform metadata to dot format if it's passed in the dict format
        if params and params.get('metadata', None):
            metadata = params.get('metadata')
            transformed_metadata = {}
            for key, value in metadata.iteritems():
                transformed_metadata['metadata.{}'.format(key)] = value
            params.pop('metadata')
            params.update(transformed_metadata)
        return params


class RequestsClient(BaseClient):
    name = 'requests'

    def request(self, method, url, headers, params=None):
        if not requests:
            raise BeyonicError(
                'requests is not installed. Please install/setup it first using pip. e.g. pip install requests>=1.0')

        kwargs = {}

        if self._verify_ssl_certs:
            kwargs['verify'] = True
        else:
            kwargs['verify'] = False

        try:
            try:
                if method.upper() == 'GET':
                    result = requests.request(method,
                                              url,
                                              headers=headers,
                                              params=params,
                                              timeout=80,
                                              **kwargs)
                else:
                    result = requests.request(method,
                                              url,
                                              headers=headers,
                                              data=self.transform_params_metadata(params),
                                              timeout=80,
                                              **kwargs)
            except TypeError, e:
                raise TypeError(
                    'Please upgrade your request library. The '
                    'underlying error was: %s' % (e,))

            content = result.content
            status_code = result.status_code
        except Exception, e:
            # Would catch just requests.exceptions.RequestException, but can
            # also raise ValueError, RuntimeError, etc.
            self._handle_request_error(e)
        return content, status_code

    def _handle_request_error(self, e):
        msg = ("Unexpected error communicating with Beyonic API.")
        err = "A %s was raised" % (type(e).__name__,)
        if str(e):
            err += " with error message %s" % (str(e),)
        else:
            err += " with no error message"
        msg = textwrap.fill(msg) + "\n\n(error: %s)" % (err,)
        raise BeyonicError(msg)


class UrlFetchClient(BaseClient):
    name = 'urlfetch'

    def request(self, method, url, headers, params=None):
        if not urlfetch:
            raise BeyonicError('urlfetch is not installed. Please install/setup it first.')
        try:
            result = urlfetch.fetch(
                url=url,
                method=method,
                headers=headers,

                validate_certificate=self._verify_ssl_certs,
                deadline=55,
                payload=self.transform_params_metadata(params)
            )
        except urlfetch.Error, e:
            self._handle_request_error(e, url)

        return result.content, result.status_code

    def _handle_request_error(self, e, url):
        if isinstance(e, urlfetch.InvalidURLError):
            msg = ("Unexpected error communicating with Beyonic API.")
        elif isinstance(e, urlfetch.DownloadError):
            msg = "There was a problem retrieving data from Beyonic."
        elif isinstance(e, urlfetch.ResponseTooLargeError):
            msg = ("Unexpected error communicating with Beyonic API.")
        else:
            msg = ("Unexpected error communicating with Beyonic API.")

        msg = textwrap.fill(msg) + "\n\n(error: " + str(e) + ")"
        raise BeyonicError(msg)


def get_default_http_client(*args, **kwargs):
    if urlfetch:
        impl = UrlFetchClient
    elif requests:
        impl = RequestsClient
    else:
        # None of them is available so let's throw an error
        raise BeyonicError('Either of requests or urlfetch is not installed. '
                           'Please install either of them using requirements.txt')

    return impl(*args, **kwargs)


"""
'API Client class interacts with api using available RequestClient or UrlFetchClient
"""


class ApiClient(object):
    """
    A client for the api
    """

    def __init__(self, api_key=None, url=None, client=None, verify_ssl_certs=True, api_version=None):
        # if not passed then let's try to get it from env variable
        if not api_key:
            try:
                api_key = os.environ['BEYONIC_ACCESS_KEY']
            except KeyError:
                raise BeyonicError('BEYONIC_ACCESS_KEY not set.')

        self._api_key = api_key
        self._api_version = api_version
        if not url:
            raise BeyonicError('Base url can\'t be empty. You should set base url using beyonic.api_endpoint_base')

        self._url = url
        self._client = client or get_default_http_client(verify_ssl_certs=verify_ssl_certs)

    def set_url(self, url):
        self._url = url

    def get(self, **kwargs):
        """
        Makes an HTTP GET request to the  API. Any keyword arguments will
        be converted to query string parameters.
        """
        return self._request("get", **kwargs)

    def post(self, **kwargs):
        """
        Makes an HTTP POST request to the  API.
        """
        return self._request("post", **kwargs)

    def put(self, **kwargs):
        """
        Makes an HTTP PUT request to the  API.
        """
        return self._request("put", **kwargs)

    def patch(self, **kwargs):
        """
        Makes an HTTP patch request to the  API.
        """
        return self._request("patch", **kwargs)

    def delete(self, **kwargs):
        """
        Makes an HTTP DELETE request to the  API.
        """
        return self._request("delete", **kwargs)

    def _request(self, method, **kwargs):
        # Check if duplicate_check_key is set in kwargs, and if so, add it to the header_overrides
        duplicate_check_key = kwargs.pop('duplicate_check_key', None)
        header_overrides = {'Duplicate-Check-Key': duplicate_check_key} if duplicate_check_key else {}

        response_content, status_code = self._client.request(
            method=method,
            url=self._url,
            headers=self._build_headers(header_overrides=header_overrides),
            params=kwargs,
        )
        return self._parse_response(response_content, status_code)

    def _build_headers(self, **kwargs):
        headers = {}
        if self._api_key:
            headers.update({"Authorization": "Token %s" % self._api_key, })

        if self._api_version:
            headers.update({"Beyonic-Version": self._api_version, })

        headers.update({"Beyonic-Client": "Python", })
        headers.update({"Beyonic-Client-Version": beyonic.__version__, })

        # Override headers
        if 'header_overrides' in kwargs:
            headers.update(kwargs['header_overrides'])

        return headers

    def _parse_response(self, response_content, status_code):
        # TODO: need exact status code for different type error
        if 200 <= status_code < 300:
            return self._value_for_response(response_content, status_code)
        else:
            raise self._exception_for_response(response_content, status_code)

    def _value_for_response(self, response_content, status_code):
        # Status_code 204 is for delete so for it returning True
        if response_content and status_code != 204:
            return GenericObject.from_json(response_content)
        else:
            return True

    def _exception_for_response(self, response_content, status_code):
        # TODO need to handle the all the status
        return ResponseError("%d error: %s" % (status_code, response_content,))



'''
' API related resources
'''
import sys, json

'''
class GenericObject(dict):
    def __init__(self, id=None, api_key=None, **params):
        super(GenericObject, self).__init__()
        self._params = params
        object.__setattr__(self, 'api_key', api_key)

        if id:
            self['id'] = id

    """
    Generic object class
    """

    @classmethod
    def from_json(cls, json_string):
        """
        Parses the given json_string
        """
        obj = json.loads(json_string, object_hook=cls)
        return obj


    def __setattr__(self, prop, val):
        if prop[0] == '_' or prop in self.__dict__:
            return super(GenericObject, self).__setattr__(prop, val)
        else:
            self[prop] = val


    def __getattr__(self, prop):
        print self
        print type(self)
        print self.keys()
        if prop in self:
            return self[prop]
        else:
            raise AttributeError

    def __delitem__(self, prop):
        # on delete, setting value to None
        if prop in self:
            self[prop] = None

    def __repr__(self):
        ident_parts = [type(self).__name__]

        if isinstance(self.get('id'), dict):
            sid = self.get('id').get('id')
            ident_parts.append('id=%s' % (sid,))
        elif isinstance(self.get('id'), basestring):
            ident_parts.append('id=%s' % (self.get('id'),))

        unicode_repr = '<%s at %s> : %s' % (
            ' '.join(ident_parts), hex(id(self)), str(self))

        print unicode_repr
        if sys.version_info[0] < 3:
            return unicode_repr.encode('utf-8')
        else:
            return unicode_repr

    def __str__(self):
        if isinstance(self.get('id'), dict):
            self = self.get('id')
        return json.dumps(self, sort_keys=True, indent=2)

'''


class GenericObject(dict):
    """
    A dict subclass that provides access to its members as if they were
    attributes.
    Note that an attribute that has the same name as one of dict's existing
    method (``keys``, ``items``, etc.) will not be accessible as an attribute.
    """

    @classmethod
    def from_json(cls, json_string):
        """
        Parses the given json_string, returning GenericObject instances instead
        of dicts.
        """
        return json.loads(json_string, object_hook=cls)

    def __setattr__(self, prop, val):
        if prop[0] == '_' or prop in self.__dict__:
            return super(GenericObject, self).__setattr__(prop, val)
        else:
            self[prop] = val

    def __getattr__(self, prop):
        """
        Provides access to the members of the dict as attributes.
        """
        if prop in self:
            return self[prop]
        else:
            raise AttributeError

    def set_client(self, client):
        self.__setattr__('_client', client)

    def save(self):
        data = self.copy()
        if '_client' in data:
            data['_client'] = None
        self._client.patch(**(data))

    def __delitem__(self, prop):
        # on delete, setting value to None
        if prop in self:
            self[prop] = None

    '''
    def __repr__(self):
        ident_parts = [type(self).__name__]

        if isinstance(self, dict):
            ident_parts.append('id=%s' % (self.get('id'),))
        elif isinstance(self.get('id'), dict):
            sid = self.get('id').get('id')
            ident_parts.append('id=%s' % (sid,))
        elif isinstance(self.get('id'), basestring):
            ident_parts.append('id=%s' % (self.get('id'),))

        unicode_repr = '<%s at %s> : %s' % (
            ' '.join(ident_parts), hex(id(self)), str(self))

        if sys.version_info[0] < 3:
            return unicode_repr.encode('utf-8')
        else:
            return unicode_repr
    '''
    
    def __str__(self):
        if isinstance(self.get('id'), dict):
            self = self.get('id')

        data = self.copy()
        if '_client' in data:
            data['_client'] = None
        return json.dumps(data, sort_keys=True, indent=2)
    


class BeyonicError(Exception):
    pass
    """
    def __init__(self, message=None, http_body=None, http_status=None, json_body=None):
        super(BeyonicError, self).__init__(message)

        self.http_body = http_body
        self.http_status = http_status
        self.json_body = json_body
    """


class ResponseError(Exception):
    """
    Raised when the Beyonic API responds with an HTTP error
    """
    pass



import os
from setuptools import setup
import os.path
import beyonic


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()


VERSION = beyonic.__version__

from os import path
from io import open
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="beyonic",
    version=VERSION,
    description="The official Python client for the Beyonic.com API",
    author="Beyonic",
    author_email="info@beyonic.com",
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=["beyonic", "beyonic.apis"],
    install_requires=["requests"],
    license="MIT",
    keywords=["api", "mobile payments", "mobile money", "beyonic", "mpesa"],
    url="https://beyonic.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



from beyonic.api_client import ApiClient
from beyonic.resources import GenericObject
from beyonic.errors import BeyonicError


class AbstractAPI(GenericObject):
    """
    AbtractApi class, all the other api class extends it
    """
    @classmethod
    def get_client(cls, client=None):
        url = cls.get_url()
        from beyonic import api_key, verify_ssl_certs, api_version

        if client:
            return ApiClient(api_key=api_key, url=url, client=client, verify_ssl_certs=verify_ssl_certs,
                             api_version=api_version)
        else:
            return ApiClient(api_key=api_key, url=url, client=None, verify_ssl_certs=verify_ssl_certs,
                             api_version=api_version)

    @classmethod
    def get_url(cls):
        from beyonic import api_endpoint_base, DEFAULT_ENDPOINT_BASE

        if not api_endpoint_base:
            api_endpoint_base = DEFAULT_ENDPOINT_BASE

        if not api_endpoint_base.endswith("/"):
            api_endpoint_base = api_endpoint_base + "/"
        cls_name = api_endpoint_base + str(cls._method_path)
        return cls_name

    @classmethod
    def list(cls, client=None, **kwargs):
        """
        This will return list of resources.
        """
        objs = cls.get_client(client).get(**kwargs)

        # Setting client object for each of the return object so that it can be used while saving data
        for obj in objs.results:
            if obj.id:
                base = cls.get_url()
                url = "%s/%s" % (base, obj.id)
                api_client = cls.get_client(client)
                api_client.set_url(url)
                obj.set_client(api_client)

        return objs

    @classmethod
    def create(cls, client=None, **kwargs):
        """
        This will create new object
        """
        return cls.get_client(client).post(**kwargs)

    @classmethod
    def get(cls, object_id, client=None, **kwargs):
        """
        This will return the single object
        """
        if not object_id:
            raise BeyonicError("Invalid ID or ID has not been specified")

        base = cls.get_url()
        url = "%s/%s" % (base, object_id)
        api_client = cls.get_client(client)
        api_client.set_url(url)
        obj = api_client.get(**kwargs)
        obj.set_client(api_client)
        return obj

    @classmethod
    def update(cls, object_id, client=None, **kwargs):
        """
        This will update the object
        """
        if not object_id:
            raise BeyonicError("Invalid ID or ID has not been specified")

        base = cls.get_url()
        url = "%s/%s" % (base, object_id)
        api_client = cls.get_client(client)
        api_client.set_url(url)
        return api_client.patch(**kwargs)

    @classmethod
    def delete(cls, object_id, client=None, **kwargs):
        """
        This will return the single object
        """
        if not object_id:
            raise BeyonicError("Invalid ID or ID has not been specified")

        base = cls.get_url()
        url = "%s/%s" % (base, object_id)
        api_client = cls.get_client(client)
        api_client.set_url(url)
        return api_client.delete(**kwargs)
